module snag.core.core;

import snag.config;
import std.format;
import std.stdio;
import std.array;
import std.process;
import std.algorithm;
import std.string;
import std.regex;

import snag.core.exception;

class Snag {
	private string[] _baseCommand;
	private SnagConfig _config;

	private bool isValidHash(string hash) {
		auto hashPattern = ctRegex!r"^[a-fA-F0-9]{7}$";
		return !matchFirst(hash, hashPattern).empty;
	}

	this(SnagConfig config) {
		_baseCommand = format(
			"git --git-dir=%s --work-tree=%s",
			config.git, config.project
		).split();
		_config = config;
	}

	private auto git(string[] command, string message, string separator = ":\n") {
		auto result = execute(_baseCommand ~ command);
		if (result.status)
			throw new SnagException(
				message ~ separator ~ result.output
			);
		return result;
	}

	void initialize() {
		git(
			["init"],
			"A Git repository initialization error occurred"
		);
		git(
			["config", "user.email", _config.email],
			"A Git repository initialization error occurred"
		);
		git(
			["config", "user.name", _config.user],
			"A Git repository initialization error occurred"
		);
		writeln(
			"The Git repository has been initialized successfully: "
			~ _config.git
		);
	}

	void status() {
		auto result = git(
			["status", "--porcelain"],
			"An error occurred while checking the file tracking status"
		);

		writeln("The following list of files requires backup:");

		result.output.split('\n').filter!(e => !e.strip.empty).each!((e) {
			writefln("\t/%s", e.strip.split[1]);
		});
	}

	void create() {
		auto result = git(
			["status", "--porcelain"],
			"An error occurred while checking the file tracking status"
		);

		if (!result.output.length)
			throw new SnagException(
				"Current file state doesn't need to be archived again"
			);

		result = execute(_baseCommand ~ ["rev-parse", "--short", "HEAD"]);
		if (result.status == 128) {
			git(["add", "."], "Failed to prepare files for archiving");
			git(["commit", "-m", "test"], "Failed to create a backup");
			writeln("Backup was created successfully");
			return;
		} else if (result.status != 0)
			throw new SnagException(
				"Failed to retrieve current snapshot information:\n"
				~ result.output
			);

		string currentSnapshot = result.output.strip('\n');

		git(
			["checkout", "-b", "mod", currentSnapshot ],
			"Failed to create a branch from the current state"
		);
		git(
			["add", "."],
			"Failed to prepare files for archiving"
		);
		git(
			["commit", "-m", "test"],
			"Failed to create a backup"
		);
		git(
			["checkout", "master"],
			"Failed to perform intermediate state switching"
		);
		git(
			["rebase", "--onto", "mod", currentSnapshot, "master", "-X", "theirs"],
			"Ошибка при rebase"
		);

		result = git(
			["rev-parse", "--short", "mod"],
			"An error occurred while checking the file tracking status"
		);

		string newSnapshot = result.output.strip('\n');

		git(
			["checkout", newSnapshot],
			"Failed to perform intermediate state switching"
		);
		git(
			["branch", "-D", "mod"],
			"Error while deleting temporary intermediate snapshot"
		);
		writeln("Backup was created successfully");
	}

	void list() {
		auto result = git(
			["rev-parse", "--short", "HEAD"],
			"Failed to retrieve current snapshot information"
		);

		string current = result.output.strip('\n');

		result = git(
			[
				"log",
				"--all",
				"--date=format:%Y.%m.%d %H:%M",
				"--pretty=format:%ad\t%h"
			],
			"Failed to retrieve the list of snapshots"
		);

		result.output.split('\n').map!(line => line.split('\t')).array
			.sort!((a, b) => a[0] > b[0]).each!(e =>
			writefln("%s\t%s\t%s", current == e[1] ? "    >" : "", e[0], e[1])
		);
	}

	void restore(string hash) {
		if (!isValidHash(hash))
			throw new SnagException(
				"Invalid snapshot hash provided"
			);

		auto result = git(
			["status", "--porcelain"],
			"An error occurred while checking the file tracking status"
		);

		if (result.output.length)
			git(
				["restore", "."],
				"Failed to reset file changes state"
			);

		git(
			["rev-parse", hash],
			"This snapshot is not available in the archive"
		);
		git(
			["checkout", hash],
			"Failed to restore the snapshot state " ~ hash
		);
		writeln("Backup was restored successfully");
	}
}
